import math
import requests
import collections
import datetime
import re
from enum import Enum
from pokefans.api.enums import *    
from pokefans.api.auth import auth
from pokefans.api.exceptions import APIException

API_BASE_URL = "https://api.pokefans.net/legacy/content"
API_CONTENT_URL = "https://api.pokefans.net/legacy/content/{id}"
API_CONTENT_BOILERPLATES = "https://api.pokefans.net/legacy/content/{id}/boilerplates"


class ContentUpdateMeta:
    """ Holds information about a content update.

    :var str username: Name of the user who updated the content.
    :var datetime.datetime datetime: Datetime the content was updated at.
    """

    def __init__(self, username, datetime):
        self.username = username
        self.datetime = datetime


class Content:
    """ Represents a content on pokefans.net.

    :var int id: The unique id of this content or ``None`` if it does not exist
        on the server.
    :var str title: The content's title
    :var str code: The content's raw code. Can contain special markup codes
        ("Inside Codes").
    :var int category: The numerical id of the content's category. The category
        ids are visible in the `Sitemap <https://pokefans.net/content/sitemap>`_
        (hover the URLs).
    :var ContentStatus status: Status of the content (published, work in
        progress, ...). Default to ``ContentStatus.IN_PROGRESS``.
    :var ContentType type: Tutorial/News/Information/... . Default to
        ``ContentType.INFORMATION``.
    :var ContentPublishMethod publish_method: Specifies how to publish this
        content (sticky on main, feature, ...). Default to
        ``ContentPublishMethod.STANDARD``.
    :var str description: A short description of the content. Is displayed on
        certain social networks.
    :var str teaser: A teaser text for this content.
    :var str notes: Internal notes for this content, not displayed publicly.
    :var str url: The content's relative URL. Changes will only be reflected
        on the server if you are a content administrator.
    :var int version: The version (revision number) of this content.
    :var str css: Special CSS formating for this content. LESS is supported.
    :var str thumbnail_url: A URL to the content's thumbnail. Will be used on
        the landing page and is displayed on certain social networks.
    :var ContentUpdateMeta create: Metadata about the creation of the content.
    :var ContentUpdateMeta update: Metadata about the last update of the content.
    """

    # Maps the JSON response content object attributes to the local attributes.
    # The tuple values read as: (attribute name, type, default value)
    jsonmap = {
        "id": ("id", int, None),
        "inhalt": ("html", str, ""),
        "inhalt_code": ("code", str, ""),
        "titel": ("title", str, ""),
        "kategorie": ("category", int, 0),
        "typ": ("type", ContentType, ContentType.INFORMATION),
        "status": ("status", ContentStatus, ContentStatus.IN_PROGRESS),
        "option_startseite": ("publish_method", ContentPublishMethod, ContentPublishMethod.STANDARD),
        "url": ("url", str, ""),
        "version": ("version", int, 1),
        "stylesheet": ("css", str, ""),
        "inhalt_beschreibung": ("description", str, ""),
        "inhalt_zusammenfassung": ("teaser", str, ""),
        "inhalt_notizen": ("notes", str, ""),
        "teaser_image_url": ("thumbnail_url", str, ""),
    }
    jsonmap_reverse = {v[0]: k for k, v in jsonmap.items()}
    _attribute_defaults = {v[0]: (v[1], v[2]) for k, v in jsonmap.items()}


    def __init__(self, *, id=None):
        """ Creates a new content object. If :param:`id` is supplied, this
        content will be fetched from the server.

        :param int id: The id of the content to load. Default to None.
        """

        self.id = id
        if id:
            self._load(id, self)
        self.cache = {}


    @classmethod
    def find(cls, *, status=None, type=None, category=None, order=None, limit=30):
        """ Returns a list of all :class:`Content` that match the parameters.

        .. warning::
            Some attributes of the :class:`Content` returned are *lazy-loaded*,
            which means that accessing them will result in an additional API
            request for loading this very data. Be vary of this when using
            lazy attributes inside a loop - you can easily run into an API
            limit!
            
            Lazy attributes for this function are:
            
            * ``code``
            * ``description``
            * ``notes``
            * ``css``

        :param ContentStatus status:
        :param ContentType type:
        :param category:
        :type category: int or list(int)
        :param Order order: Ascending or descending ordering (by id).
        :param int limit: Specifies the size of the returned list. Default to 30.
        """

        filter_params = {}
        if status:
            filter_params["filter_status"] = status.value
        if type:
            filter_params["filter_typ"] = type.value
        if category:
            if isinstance(category, list) or isinstance(category, tuple):
                filter_params["filter_kategorie[]"] = category
            else:
                filter_params["filter_kategorie"] = category
        if order:
            filter_params["order"] = order.value
            filter_params["order_by"] = "id"
        
        r = requests.get(
            API_BASE_URL,
            params=filter_params,
            auth=auth
        )
        r.raise_for_status()
        data = r.json()
        contents = [cls._from_json(c) for c in data]
        
        # Load all other pages, if there are any
        results_per_page = 30
        last_page = min(int(r.headers["X-Total-Pages"]), math.ceil(limit / results_per_page))
        for page in range(2, last_page + 1):
            filter_params["page"] = page
            r = requests.get(
                API_BASE_URL,
                params=filter_params,
                auth=auth
            )
            r.raise_for_status()
            data = r.json()
            contents += [cls._from_json(c) for c in data]
        
        return contents[:limit]


    @classmethod
    def _load(cls, id, content=None):
        """ Loads the content with the given :param:`id`. If :param:`content` is
        supplied, it will be populated instead of a new object.

        :param int id: The id of the content to load.
        :param Content content: The content to populate. Default to ``None``.
        """

        if not content:
            content = cls()

        r = requests.get(API_CONTENT_URL.format(id=id), auth=auth)
        r.raise_for_status()
        cls._from_json(r.json(), content)


    @classmethod
    def _from_json(cls, json, content=None):
        """ Populates a :class:`Content` with the data provided by the API.

        :param str json: The JSON data provided by the API.
        :param Content content: An instance of a content to populate.
        """

        if not content:
            content = cls()
        
        # Legacy cleanups
        json["typ"] = json["typ"].replace("artikel", "spielhilfe")
        json["typ"] = json["typ"].replace("inhalt", "spielhilfe")

        for attr, value in json.items():
            try:
                attr_local, attr_type, attr_default = cls.jsonmap[attr]
                setattr(content, attr_local, attr_type(value))
            except KeyError:
                continue
        
        content.create = ContentUpdateMeta(
            json["creation"]["username"],
            datetime.datetime.fromtimestamp(int(json["creation"]["timestamp"]))
        )
        content.update = ContentUpdateMeta(
            json["update"]["username"],
            datetime.datetime.fromtimestamp(int(json["update"]["timestamp"]))
        )

        return content
    

    def save(self, *, squash=False):
        """ Stores this content on the server. If the content was newly created,
        the ``id`` of this object will be updated.

        :param bool squash: Eligible users can prevent versioning of
            contents. If `squash` is set to true despite having the
            permissions, no error will be raised.
        """

        update_data = {}
        for post_name, (attr_name, *others) in self.jsonmap.items():
            # We can't use hasattr(), because it internally does a getattr(),
            # which could lead to an additional unneccessary API request
            if attr_name in self.__dict__:
                value = getattr(self, attr_name)
                if isinstance(value, Enum):
                    value = value.value
                update_data[post_name] = str(value)
        
        if squash:
            update_data["squash"] = 1

        if self.id:
            update_url = API_CONTENT_URL.format(id=self.id)
        else:
            update_url = API_BASE_URL

        r = requests.post(update_url, data=update_data, auth=auth)
        r.raise_for_status()
        data = r.json()
        self.id = int(data["id"])
    

    def update_block(self, block_id, value):
        """ Updates the block with the given id in this content.

        :param str block_id: The id of the :class:`Block` to update.
        :param str value: The new value for the block's code.
        """

        block = Block(id=block_id)
        block.code = value
        self.code = block.replace_in(self.code)


    @property
    def boilerplates(self):
        """ Returns all :class:`Boilerplate` in this content.

        :return: The :class:`Boilerplate` for this content.
        :rtype: tuple(Boilerplate)
        """
        
        if not self.id:
            return []
        
        r = requests.get(API_CONTENT_BOILERPLATES.format(id=self.id), auth=auth)
        r.raise_for_status()
        boilerplates_data = r.json()

        boilerplates = []
        for bp in boilerplates_data:
            boilerplate = Boilerplate(placeholder=bp["platzhalter"].replace("%", ""), id=int(bp["id"]))
            boilerplate.title = bp["titel"]
            boilerplates.append(boilerplate)

        return tuple(boilerplates)
    

    def add_boilerplate(self, boilerplate):
        """ Adds a :class:`Boilerplate` to this content. This method can only be
        called on saved contents and saved boilerplates.

        :param Boilerplate boilerplate:
        :raises APIException: If this content or the boilerplate is not saved.
        """

        if not self.id:
            raise APIException("Boilerplates can only be added to saved contents. save() this content before calling this method.")
        if not boilerplate.id:
            raise APIException("Only existing boilerplates can be added to a content. save() the boilerplate before calling this method.")
        
        r = requests.post(
            API_CONTENT_BOILERPLATES.format(id=self.id),
            data={"textbaustein_id": boilerplate.id, "platzhalter": boilerplate.placeholder},
            auth=auth
        )
        r.raise_for_status()
    

    @property
    def blocks(self):
        """ A dictionary of all :class:`Block` in the content's code. The
        dictionary maps the block id to the :class:`Block` instance.

        :return: The :class:`Block` in ``Content.code``.
        :rtype: dict(str, Block)
        """

        return Block.load_from(self.code)
    

    def reload(self):
        """ Reloads this content from the server. All unsaved local changes
        will be discarded. If this content does not exist on the server, nothing
        happens. """

        if self.id:
            self._load(self.id, self)
    

    def __getattr__(self, name):
        try:
            attrtype, default_value = self._attribute_defaults[name]
        except KeyError:
            raise AttributeError()

        if not self.id:
            setattr(self, name, default_value)
        else:
            if not self.cache:
                r = requests.get(API_CONTENT_URL.format(id=self.id), auth=auth)
                r.raise_for_status()
                self.cache = r.json()
            setattr(self, name, self.cache[self.jsonmap_reverse[name]])
        return getattr(self, name)
    

    def __str__(self):
        return self.code
    

    def __repr__(self):
        return '<Content id={id}>'.format(id=self.id)
    

class Boilerplate(Content):
    """ A Boilerplate is a :class:`Content` that is included in another
    :class:`Content`. Boilerplates feature a ``placeholder``, which gets
    inserted into the content's code. When displaying this content, the
    placeholder is replaced with its actual content.

    When converting a :class:`Boilerplate` into a string, it returns its
    placeholder. That makes it possible to use a :class:`Boilerplate` object
    directly in a content code string::

        content = Content(id=1234)
        boilerplate = Boilerplate(placeholder="YOUR_PLACEHOLDER")
        boilerplate.code = "Boilerplate content"
        boilerplate.save()

        content.code = "Boilerplate include: " + str(boilerplate)
        content.add_boilerplate(boilerplate)

        print(content.code)
        # "Boilerplate include: %YOUR_PLACEHOLDER%"
    

    .. warning::
        *All* attributes except ``placeholder``, ``type`` and ``title``
        are lazy-loaded. This means that accessing any other attribute will
        result in an additional API request for loading this data. Be vary of
        this, as you can easily run into an API limit.

    :var str placeholder: A placeholder for this boilerplate in the content.
        Must only contain alphanumerical characters and underscores.
    """

    def __init__(self, *, placeholder, id=None):
        """ Initializes a new object.

        :param str placeholder: The placeholder for this Boilerplate. Must be a
            string containing only alphanumerical characters and underscores.
        :param int id: The id of the boilerplate content.
        :raises ValueError: If an invalid placeholder was supplied.
        """

        if not re.match(r"^[a-zA-Z0-9\_]+$", placeholder):
            raise ValueError("Invalid placeholder supplied. Only alphanumerical characters and underscores are supported.")
        
        self.id = id
        self.placeholder = placeholder
        self.type = ContentType.BOILERPLATE
    

    def __str__(self):
        return "%{}%".format(self.placeholder)
    

    def __repr__(self):
        return '<Boilerplate id={id} placeholder="{placeholder}" title="{title}">'.format(id=self.id, placeholder=self.placeholder, title=self.title)


class Block:
    """ A Block contains HTML code. Think of a block as a named section of HTML
    code in a :class:`Content` you can find and update easily. See for example::

        content = Content(id=1234)
        block = Block(id="lukeskywalker/myscript.py")
        block.content = "content"
        content.code = "Here is a block: " + str(block)
        print(content.code)
        # "Here is a block: <!-- BEGIN lukeskywalker/myscript.py -->[...] content<!-- END lukeskywalker/myscript.py -->"
        
        content.update_block("lukeskywalker/myscript.py", "new content")
        print(content.code)
        # "Here is a block: <!-- BEGIN lukeskywalker/myscript.py -->[...] new content<!-- END lukeskywalker/myscript.py -->"
    
    :var str id: A unique identifier for this block. This can be any string that
        does not contain whitespaces. However, you should use a recognizable
        format, containing your username and some information about the script.
        For example ``lukeskywalker/places.py:myfancytable``.
    :var str code: The content of the block.
    """

    BLOCK_REGEXP = "<!-- BEGIN {id} -->(.*)<!-- END {id} -->"

    def __init__(self, *, id):
        """ Initializes a new object.

        :param int id: The id of this block. Must not contain whitespaces.
        :raises ValueError: If the format of ``id`` is invalid.
        """

        if not re.match(r"^[^\s]+$", id):
            raise ValueError("A block's id must not contain whitespaces: " + str(id))
        self.id = id
        self.code = ""
        self.regexp = re.compile(self.BLOCK_REGEXP.format(id=re.escape(id)), re.DOTALL)
    

    def replace_in(self, s):
        """ Replaces the code of the blocks in ``s`` having the same ``id`` as
        this block with the code of this block.

        :param str s: The string to replace the blocks in.
        :rtype: str
        """

        return self.regexp.sub(str(self), s)
    

    @classmethod
    def load_from(cls, s):
        """ Returns all :class:`Block` in the string provided.

        :param str s: The string to extract the blocks from.
        :return: A dictionary of all :class:`Block` in the string. Keys are the
            block ids, values are instances of :class:`Block`.
        :rtype: dict(str, Block)
        """
        
        blocks = {}
        for blockdata in re.findall(cls.BLOCK_REGEXP.format(id="([^\s]+)"), s, re.DOTALL):
            block = cls(id=blockdata[0].strip())
            block.code = blockdata[1].strip()
            blocks[block.id] = block
        return blocks


    def __str__(self):
        out = """<!-- BEGIN {id} -->
    {code}
<!-- END {id} -->""".format(id=self.id, code=self.code)
        return out
    

    def __repr__(self):
        return '<Block id="{}">'.format(self.id)
